/*
 * $Id: bogons.c,v 1.3 2004/10/24 00:06:38 cbiere Exp $
 *
 * Copyright (c) 2004, Raphael Manfredi
 *
 *----------------------------------------------------------------------
 * This file is part of gtk-gnutella.
 *
 *  gtk-gnutella is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  gtk-gnutella is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gtk-gnutella; if not, write to the Free Software
 *  Foundation, Inc.:
 *      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------
 */

/**
 * @file
 *
 * Support for IP bogons detection.
 */

#include "common.h"

RCSID("$Id: bogons.c,v 1.3 2004/10/24 00:06:38 cbiere Exp $");

#include "bogons.h"
#include "settings.h"

#include "lib/file.h"
#include "lib/misc.h"
#include "lib/glib-missing.h"
#include "lib/iprange.h"
#include "lib/walloc.h"
#include "lib/watcher.h"

#include "if/gnet_property_priv.h"
#include "if/bridge/c2ui.h"

#include "lib/override.h"		/* Must be the last header included */

#define THERE	GUINT_TO_POINTER(0x2)

static const gchar bogons_file[] = "bogons.txt";
static const gchar bogons_what[] = "Bogus IP addresses";

static gpointer bogons_db;		/* The database of bogus CIDR ranges */

/**
 * Load bogons data from the supplied FILE.
 * Returns the amount of entries loaded.
 */
static gint
bogons_load(FILE *f)
{
	gchar line[1024];
	gchar *p;
	guint32 ip, netmask;
	int linenum = 0;
	gint count = 0;
	gint bits;
	iprange_err_t error;

	bogons_db = iprange_make(NULL, NULL);

	while (fgets(line, sizeof(line), f)) {
		linenum++;
		if (*line == '\0' || *line == '#')
			continue;

		/*
		 * Remove all trailing spaces in string.
		 * Otherwise, lines which contain only spaces would cause a warning.
		 */
	
		p = strchr(line, '\0');	
		while (--p >= line) {
			guchar c = (guchar) *p;
			if (!is_ascii_space(c))
				break;
			*p = '\0';
		}
		if ('\0' == *line)
			continue;

		if (!gchar_to_ip_and_mask(line, &ip, &netmask)) {
			g_warning("%s, line %d: invalid IP or netmask \"%s\"",
				bogons_file, linenum, line);
			continue;
		}

		bits = 32;
		while (0 == (netmask & 0x1)) {
			netmask >>= 1;
			bits--;
		}

		error = iprange_add_cidr(bogons_db, ip, bits, THERE);

		switch (error) {
		case IPR_ERR_OK:
			break;
		case IPR_ERR_RANGE_OVERLAP:
			error = iprange_add_cidr_force(bogons_db, ip, bits, THERE, NULL);
			if (error == IPR_ERR_OK) {
				g_warning("%s, line %d: "
					"entry \"%s\" (%s/%d) superseded earlier smaller range",
					bogons_file, linenum, line, ip_to_gchar(ip), bits);
				break;
			}
			/* FALL THROUGH */
		default:
			g_warning("%s, line %d: rejected entry \"%s\" (%s/%d): %s",
				bogons_file, linenum, line, ip_to_gchar(ip), bits,
				iprange_strerror(error));
			continue;
		}

		count++;
	}

	if (dbg) {
		iprange_stats_t stats;

		iprange_get_stats(bogons_db, &stats);

		g_message("loaded %d bogus IP ranges", count);
		g_message("bogons stats: count=%d level2=%d heads=%d enlisted=%d",
			stats.count, stats.level2, stats.heads, stats.enlisted);
	}

	return count;
}

/**
 * Watcher callback, invoked when the file from which we read the bogus
 * addresses changed.
 */
static void
bogons_changed(const gchar *filename, gpointer unused_udata)
{
	FILE *f;
	gchar buf[80];
	gint count;

	(void) unused_udata;

	f = file_fopen(filename, "r");
	if (f == NULL)
		return;

	bogons_close();
	count = bogons_load(f);

	gm_snprintf(buf, sizeof(buf), "Reloaded %d bogus IP ranges.", count);
	gcu_statusbar_message(buf);
}

/**
 * Loads the bogons.txt into memory, choosing the first file we find
 * among the several places we look at, typically:
 *
 *    ~/.gtk-gnutella/bogons.txt
 *    /usr/share/gtk-gnutella/bogons.txt
 *    /home/src/gtk-gnutella/bogons.txt
 *
 * The selected file will then be monitored and a reloading will occur
 * shortly after a modification.
 */
static void
bogons_retrieve(void)
{
	FILE *f;
	gint idx;
	gchar *filename;
	file_path_t fp[] = {
		{ settings_config_dir(), bogons_file },
		{ PRIVLIB_EXP, bogons_file },
#ifndef OFFICIAL_BUILD 
		{ PACKAGE_SOURCE_DIR, bogons_file }
#endif
	};

	f = file_config_open_read_norename_chosen(
			bogons_what, fp, G_N_ELEMENTS(fp), &idx);

	if (!f)
	   return;

	filename = make_pathname(fp[idx].dir, fp[idx].name);
	watcher_register(filename, bogons_changed, NULL);
	G_FREE_NULL(filename);

	bogons_load(f);
}

/**
 * Called on startup. Loads the bogons.txt into memory.
 */
void
bogons_init(void)
{
	bogons_retrieve();
}

/**
 * Frees all entries in the hostiles
 */
void
bogons_close(void)
{
	iprange_free_each(bogons_db, NULL);
	bogons_db = NULL;
}

/**
 * Check the given IP agains the entries in the bogus IP database.
 * Returns TRUE if found, and FALSE if not.
 */
gboolean
bogons_check(guint32 ip)
{
	return THERE == iprange_get(bogons_db, ip);
}

/* vi: set ts=4 sw=4 cindent: */
